#!/usr/bin/env python
import imp, os, sys, shutil, subprocess, tempfile, time, string

from fabric.api import env, run, local, cd
from fabric.operations import put

# TODO the ubuntu vm is used with the 'openerp' user. A different user
# should be used as 'openerp' is also the user running openerp-server.
# TODO Test the running server after the .deb is installed.

#
# Configuration. Instead of modifying the following variables, you can simply
# create a local-config.py file.
#

# Bazaar repositories to package.
REPOS = os.path.realpath('repos')

# Where the resulting .tar.gz, .exe, and .deb will be generated.
PACKAGES = os.path.realpath('packages')

# Where the log files will be written.
RESULTS = os.path.realpath('results')

WINDOWS_XP_SNAPSHOT = '/home/odoo/vm/winxpsp3_openerp.vmdk'
WINDOWS_XP_IDENTITY = '-i /home/odoo/vm/buildbot_rsa'

UBUNTU_SNAPSHOT = '/home/odoo/vm/ubuntu-10.04.2-server-i386.img'
UBUNTU_IP = '127.0.0.1'
UBUNTU_PORT = '2225'
UBUNTU_IDENTITY = '-i /home/odoo/vm/ubuntu-10.04.2-server-i386_rsa'

# This is used to derive minor and major versions for the Windows packaging.
# This could be part of the PROJECTS dictionary below.
VERSIONS = {
    '6.0': {
        'major': '6',
        'minor': '0',
    },
    '6.1-dev': {
        'major': '6',
        'minor': '1-dev',
    }
}

# For each branch we want to package, for each project, we give:
# - the path to the setup.py if any,
# - the directory where all the code is (actually used only in
#   the server to know the location of the addons directory),
# - the path to the release.py file,
# - the repository URI.
PROJECTS = {
    '6.0': {
        'addons': (None,       'openerp',    'release.py', 'http://bazaar.launchpad.net/~openerp/openobject-addons/6.0'),
        'server': ('setup.py', 'bin',        'release.py', 'http://bazaar.launchpad.net/~openerp/openobject-server/6.0'),
        'client': ('setup.py', 'bin',        'release.py', 'http://bazaar.launchpad.net/~openerp/openobject-client/6.0'),
        'web':    ('setup.py', 'openobject', 'release.py', 'http://bazaar.launchpad.net/~openerp/openobject-client-web/6.0'),
    },
    '6.1-dev': {
        'addons': (None,       'openerp',    'release.py', 'http://bazaar.launchpad.net/~openerp/openobject-addons/trunk'),
        'server': ('setup.py', 'openerp',    'release.py', 'http://bazaar.launchpad.net/~openerp/openobject-server/trunk'),
        'client': ('setup.py', 'bin',        'release.py', 'http://bazaar.launchpad.net/~openerp/openobject-client/trunk'),
        'web':    ('setup.py', 'openobject', 'release.py', 'http://bazaar.launchpad.net/~openerp/openobject-client-web/trunk'),
    },
}

# Replaces configuration variables from user-defined file.
configurables = [ 'REPOS', 'PACKAGES', 'RESULTS',
  'WINDOWS_XP_SNAPSHOT', 'WINDOWS_XP_IDENTITY',
  'UBUNTU_SNAPSHOT', 'UBUNTU_IDENTITY',
  'VERSIONS', 'PROJECTS'
]
if os.path.exists('local-conf.py'):
    conf = {}
    execfile('local-conf.py', conf)
    def use(name):
        if name in conf:
            globals()[name] = conf.get(name)
    for name in configurables:
        use(name)

def show_configurables():
    """ Show the name and value of the configuration variables """
    for name in configurables:
        print string.ljust(name,20), ":", globals()[name]

def server_setup_exe_name(version,revno,date):
    return "openerp-server-setup-%s.%s-%s.exe" % (version,revno,date)

def server_setup_exe_path(version,revno,date):
    return "packages/%s/openerp-server-setup-%s.%s-%s.exe" % (version,version,revno,date)

def current_project_tars(project, version):
    directory = os.path.abspath(os.path.join(REPOS,version,project))
    cmd = ['python', os.path.join(directory, PROJECTS[version][project][0]), '--fullname']
    name = subprocess.Popen(cmd, cwd=directory, stdout=subprocess.PIPE).communicate()[0].strip()
    date = time.strftime("%Y%m%d")
    revno = subprocess.Popen(['bzr','revno',os.path.join(REPOS,version,project)], stdout=subprocess.PIPE).communicate()[0].strip()
    return ("%s.tar.gz" % name, "%s-%s-r%s.tar.gz" % (name, date, revno))

def create_directory(dirname):
    if not os.path.exists(dirname):
        os.mkdir(dirname)

create_directory(REPOS)
create_directory(PACKAGES)
create_directory(RESULTS)
for version in PROJECTS:
    create_directory(os.path.join(REPOS,version))
    create_directory(os.path.join(PACKAGES,version))

# unused
DIR_WEBSITE = os.path.join('/', 'var', 'www', 'www.openerp.com', 'www', 'download')
DIR_SOURCE = os.path.join(DIR_WEBSITE, 'unstable', 'source')
DIR_WIN32 = os.path.join(DIR_WEBSITE, 'unstable', 'win32')

env.hosts = ['root@openerp.com']
env.show = ['debug']

def system(l):
    print l
    if isinstance(l,list):
        rc=os.spawnvp(os.P_WAIT, l[0], l)
    elif isinstance(l,str):
        tmp=['sh','-c',l]
        rc=os.spawnvp(os.P_WAIT, tmp[0], tmp)
    return rc

def versions():
    """ Show the possible versions """
    for version in PROJECTS:
        print version

def update_repos():
    """ Make bzr pull on each branch and copy the addons in the server """
    #system("[ -d .bzr ] || bzr init-repository .")
    for version in PROJECTS:
        for project in PROJECTS[version]:
            dest = os.path.join(REPOS,version,project)
            if not os.path.isdir(dest):
                system(['bzr','branch',PROJECTS[version][project][3],dest])
            else:
                system(['bzr','pull','-d',dest,PROJECTS[version][project][3]])
        cmd = "rsync -av --delete"
        cmd += " --exclude .bzr/"
        cmd += " --exclude .bzrignore"
        cmd += " --exclude /__init__.py"
        cmd += " --exclude /base"
        cmd += " --exclude /base_quality_interrogation.py"
        cmd += " %s/%s/addons/" % (REPOS,version)
        cmd += " %s/%s/server/%s/addons/" % (REPOS,version, PROJECTS[version]['server'][1])
        system(cmd)

def sdist(version, project):
    """ Generate the source tarball for specific version/project"""
    if PROJECTS[version][project][0]: # if there is a setup.py
        packages_dir = os.path.abspath(os.path.join(PACKAGES,version))
        directory = os.path.abspath(os.path.join(REPOS,version,project))
        cmd=['python', os.path.join(directory, PROJECTS[version][project][0]), '--quiet', 'sdist', '-d', packages_dir]
        print cmd
        subprocess.Popen(cmd, cwd=directory).communicate()
        name, fullname = current_project_tars(project, version)
        print "rename", os.path.join(packages_dir, name), os.path.join(packages_dir, fullname)
        os.rename(os.path.join(packages_dir, name), os.path.join(packages_dir, fullname))

def sdist_all():
    """ Generate the source tarballs for all versions/projects """
    for version in PROJECTS:
        for project in PROJECTS[version]:
            sdist(version, project)

def debian_build(version=None):
    """ Generate the .deb """
    if version is None or version not in PROJECTS:
        print "debian_build: version should be one of"
        for v in PROJECTS:
            print "-", v
        sys.exit()
    name, fullname = current_project_tars('server', version)
    print "Creating .deb for", fullname
    if not os.path.exists('packages/%s/%s' % (version, fullname)):
        print "packages/%s/%s doesn't exist, exiting," % (version, fullname)
        print "you can generate it by running fab sdist:%s,server" % version
        sys.exit()
    # cp tar.gz to debianization, untar it, cd it and run dpkg-buildpackage, retreive resulting .deb.
    system('rm -rf debianization')
    system('mkdir debianization')
    system('cp -a packages/%s/%s debianization/' % (version, fullname))
    system('tar xvf debianization/%s --directory debianization' % fullname)
    subprocess.Popen(['dpkg-buildpackage', '-rfakeroot'], cwd='debianization/openerp-server-%s'%version).communicate()
    system('mv debianization/openerp-server_6.0.1-1_all.deb packages/%s/openerp-server-%s.deb' % (version,version)) # TODO correct version
    system('rm -rf debianization')

def debian_kvm(work):
    l=("kvm -nographic -net nic -net user,hostfwd=tcp:127.0.0.1:%s-:22 -snapshot %s"%(UBUNTU_PORT, UBUNTU_SNAPSHOT)).split(" ")
    print l
    pid=os.spawnvp(os.P_NOWAIT, l[0], l)
    print "sleep 5"
    time.sleep(5)
    work('openerp@%s' % UBUNTU_IP, UBUNTU_PORT) # TODO config variable for the user
    print "kill",pid
    #os.kill(pid,15)

def debian_test(version):
    """ Install a .deb in a Ubuntu virtual machine """
    def work(ip, port):
        system("scp %s -P %s packages/%s/openerp-server-%s.deb %s:" % (UBUNTU_IDENTITY,port,version,version,ip))
        system('ssh %s -p %s %s "sudo dpkg -i openerp-server-%s.deb"' % (UBUNTU_IDENTITY,port,ip,version))
        system('ssh %s -p %s %s "sudo apt-get --yes -f install"' % (UBUNTU_IDENTITY,port,ip))
        system('ssh %s -p %s %s "createdb pack"' % (UBUNTU_IDENTITY,port,ip))
        # Create the config file. TODO It seems there is a bug when there is no config file.
        system('ssh %s -p %s %s "openerp-server -s --stop-after-init"' % (UBUNTU_IDENTITY,port,ip))
        system('ssh %s -p %s %s "openerp-server -d pack -i base,sale --stop-after-init --log-level=test --logfile=openerp-server.log"' % (UBUNTU_IDENTITY,port,ip))
        system("scp %s -P %s %s:openerp-server.log %s/openerp-server-%s-deb.log"%(UBUNTU_IDENTITY,port,ip,RESULTS,version))
    debian_kvm(work)

def windows_build(version,ip,port):
    # The revno is the one from the server
    revno = subprocess.Popen(['bzr','revno',os.path.join(REPOS,version,'server')], stdout=subprocess.PIPE).communicate()[0].strip()
    date = time.strftime("%Y%m%d")
    system('rsync -av -e "ssh %s -p %s" --delete --exclude .bzr/ --exclude .bzrignore --exclude /packages/ %s/%s/ %s:openerp-packaging/'%(WINDOWS_XP_IDENTITY,port,REPOS,version,ip))
    f = open('windows-installer/Makefile.version','w')
    f.write("MAJOR_VERSION=%s\n" % VERSIONS[version]['major'])
    f.write("MINOR_VERSION=%s\n" % VERSIONS[version]['minor'])
    f.write("REVISION_VERSION=%s\n" % revno)
    f.write("BUILD_VERSION=%s\n" % date)
    f.close()
    system('rsync -av -e "ssh %s -p %s" --delete windows-installer %s:openerp-packaging/'%(WINDOWS_XP_IDENTITY,port,ip))
    system('ssh %s -p %s %s "cd openerp-packaging/windows-installer;time make allinone;"'%(WINDOWS_XP_IDENTITY,port,ip))
    system('rsync -av -e "ssh %s -p %s" %s:openerp-packaging/windows-installer/files/ %s/%s/ '%(WINDOWS_XP_IDENTITY,port,ip,PACKAGES,version))

def windows_vmare():
    windows_build('Stef@172.16.225.128','22')

def windows_kvm(version):
    """ Generate a Windows installer usgin a Windows virtual machine """
    l=("kvm -vnc 127.0.0.1:1 -net nic -net user,hostfwd=tcp:127.0.0.1:2222-:22 -snapshot %s"%WINDOWS_XP_SNAPSHOT).split(" ")
    print l
    pid=os.spawnvp(os.P_NOWAIT, l[0], l)
    print "sleep 60"
    time.sleep(60)
    windows_build(version,'Stef@127.0.0.1','2222')
    print "kill",pid
    os.kill(pid,15)

def windows_test_base(version,pgpwd):
    """ Install a Windows .exe in a Windows virtual machine """
    # The revno is the one from the server
    revno = subprocess.Popen(['bzr','revno',os.path.join(REPOS,version,'server')], stdout=subprocess.PIPE).communicate()[0].strip()
    date = time.strftime("%Y%m%d")
    # Could use more the fabric stuffs instead.
    ip = 'Stef@127.0.0.1'
    port = '2223'
    l=("kvm -vnc 127.0.0.1:2 -net nic -net user,hostfwd=tcp:127.0.0.1:2223-:22,hostfwd=tcp:127.0.0.1:5433-:5432 -snapshot %s"%WINDOWS_XP_SNAPSHOT).split(" ")
    print l
    pid=os.spawnvp(os.P_NOWAIT, l[0], l)
    print "sleep 60"
    time.sleep(60)
    # Change postgres config so it accepts our connections.
    system("scp %s -P %s pg_hba.conf '%s:/cygdrive/c/Program\ Files/OpenERP\ 6.0/PostgreSQL/data/pg_hba.conf'"%(WINDOWS_XP_IDENTITY,port,ip))
    system("scp %s -P %s postgresql.conf '%s:/cygdrive/c/Program\ Files/OpenERP\ 6.0/PostgreSQL/data/postgresql.conf'"%(WINDOWS_XP_IDENTITY,port,ip))
    system("""ssh %s -p %s %s 'net stop "PostgreSQL For OpenERP"'"""%(WINDOWS_XP_IDENTITY,port,ip))
    system("""ssh %s -p %s %s 'net start "PostgreSQL For OpenERP"'"""%(WINDOWS_XP_IDENTITY,port,ip))
    system("scp %s -P %s packages/%s/openerp-server-setup-%s.%s-%s.exe %s:"%(WINDOWS_XP_IDENTITY,port,version,version,revno,date,ip))
    system('ssh %s -p %s %s "./openerp-server-setup-%s.%s-%s.exe /S /D=\'c:\\cygwin\\home\\Stef\\server\'"'%(WINDOWS_XP_IDENTITY,port,ip,version,revno,date))
    subprocess.call('createdb -e -h 127.0.0.1 -p 5433 -U openpg pack'.split(' '), env={'PGPASSWORD':pgpwd})
    system('ssh %s -p %s %s "./server/openerp-server.exe -d pack -i base,report_webkit --stop-after-init --log-level=test"'%(WINDOWS_XP_IDENTITY,port,ip))
    system("scp %s -P %s %s:server/openerp-server.log %s/openerp-server-%s-r%s-%s.log"%(WINDOWS_XP_IDENTITY,port,ip,RESULTS,version,revno,date))
    print "kill",pid
    os.kill(pid,15)

def windows_test_setup_base(setuppath,pgpwd):
    setupdir, setupexe = os.path.split(setuppath)
    setup_no_ext, ext = os.path.splitext(setupexe)
    # Could use more the fabric stuffs instead.
    ip = 'Stef@127.0.0.1'
    port = '2224'
    l=("kvm -vnc 127.0.0.1:3 -net nic -net user,hostfwd=tcp:127.0.0.1:2224-:22,hostfwd=tcp:127.0.0.1:5433-:5432 -snapshot %s"%WINDOWS_XP_SNAPSHOT).split(" ")
    print l
    pid=os.spawnvp(os.P_NOWAIT, l[0], l)
    print "sleep 60"
    time.sleep(60)
    # Change postgres config so it accepts our connections.
    system("scp %s -P %s pg_hba.conf '%s:/cygdrive/c/Program\ Files/OpenERP\ 6.0/PostgreSQL/data/pg_hba.conf'"%(WINDOWS_XP_IDENTITY,port,ip))
    system("scp %s -P %s postgresql.conf '%s:/cygdrive/c/Program\ Files/OpenERP\ 6.0/PostgreSQL/data/postgresql.conf'"%(WINDOWS_XP_IDENTITY,port,ip))
    system("""ssh %s -p %s %s 'net stop "PostgreSQL For OpenERP"'"""%(WINDOWS_XP_IDENTITY,port,ip))
    system("""ssh %s -p %s %s 'net start "PostgreSQL For OpenERP"'"""%(WINDOWS_XP_IDENTITY,port,ip))
    system("scp %s -P %s %s %s:"%(WINDOWS_XP_IDENTITY,port,setuppath,ip))
    system('ssh %s -p %s %s "./%s /S /D=\'c:\\cygwin\\home\\Stef\\server\'"'%(WINDOWS_XP_IDENTITY,port,ip,setupexe))
    subprocess.call('createdb -e -h 127.0.0.1 -p 5433 -U openpg pack'.split(' '), env={'PGPASSWORD':pgpwd})
    system('ssh %s -p %s %s "./server/openerp-server.exe -d pack -i base,report_webkit --stop-after-init --log-level=test"'%(WINDOWS_XP_IDENTITY,port,ip))
    system("scp %s -P %s %s:server/openerp-server.log %s/%s.log"%(WINDOWS_XP_IDENTITY,port,ip,RESULTS,setup_no_ext))
    print "kill",pid
    os.kill(pid,15)

def upload_tar():
    """ Upload the tarballs (source) """
    #run('mkdir %s -p' % DIR_SOURCE)
    #put(os.path.join(PACKAGES, GLOB_TARBALL), DIR_SOURCE)

def upload_win32():
    """ Upload the Windows Installers """
    #run('mkdir %s -p' % DIR_WIN32)
    #put(os.path.join(PACKAGES, GLOB_WIN32), DIR_WIN32)

def upload():
    """ Upload the Windows Installer and the tarballs """
    upload_tar()
    upload_win32()

def make_all():
    daily()
    upload()

def daily(pgpwd):
    update_repos()
    sdist_all()
    for version in PROJECTS:
        windows_kvm(version)
        windows_test_base(version,pgpwd)

